"""Finds plateau region by segmented regression."""

import numpy as np

__all__ = [
    "plateau_type2",
    "plateau_type3",
]


def _ols(Xi, Y):
    XT_X_inv = np.linalg.inv(Xi.T @ Xi)
    params = XT_X_inv @ (Xi.T @ Y)
    return params


def _segreg(x, Y, psi0, tol=1e-5, maxiter=30):
    r"""Segmented regression with one breakpoint.

    Parameters
    ----------
    x, Y : (M,) ndarray
        Data points.
    psi0 : scalar
        Initial guess for breakpoint coordinate.
    tol : float, default=1e-5
        Convergence tolerance.
    maxiter : int, default=30
        Force break after this iterations.

    Returns
    -------
    params : (4,) ndarray
        Estimated parameters: b0, b1, b2, psi.
    reached_max : bool
        Iteration is finished not by convergence but by reaching maximum iteration.
    """
    Xi = np.array(
        [
            np.ones_like(x),
            x,
            (x - psi0) * np.heaviside(x - psi0, 0),
            -np.heaviside(x - psi0, 0),
        ]
    ).T

    b0, b1, b2, gamma = _ols(Xi, Y)
    RSS = np.sum((Y - _segreg_predict(x, b0, b1, b2, psi0)) ** 2)

    psi_converged = False
    for _ in range(maxiter):
        RSS_new = RSS
        lamda = 1
        while True:
            psi0_new = psi0 + lamda * gamma / b2
            RSS_new = np.sum((Y - _segreg_predict(x, b0, b1, b2, psi0_new)) ** 2)
            lamda /= 2

            if (psi0_new <= x[0]) or (psi0_new >= x[-1]):
                # exceeded domain; make step size smaller
                continue
            if RSS_new >= RSS:
                # RSS not decreased; make step size smaller
                continue
            psi_converged = np.abs(psi0 - psi0_new) <= tol
            if psi_converged:
                break

        if not psi_converged:
            psi_converged = np.abs(psi0 - psi0_new) <= tol
        if psi_converged:
            psi0 = psi0_new
            reached_max = False
            break

        psi0 = psi0_new
        RSS = RSS_new
        Xi[:, 2] = (x - psi0) * np.heaviside(x - psi0, 0)
        Xi[:, 3] = -np.heaviside(x - psi0, 0)
        b0, b1, b2, gamma = _ols(Xi, Y)
    else:
        reached_max = True

    params = np.array([b0, b1, b2, psi0_new])
    return params, reached_max


def _segreg_predict(x, b0, b1, b2, psi):
    x = np.asarray(x)
    return b0 + b1 * x + b2 * (x - psi) * np.heaviside(x - psi, 0)


def plateau_type2(x, Ys, peaks, knees):
    """Find plateau for type 2 heavy edge profiles.

    Parameters
    ----------
    x : array of shape (M,)
        X grid of profiles.
    Ys : array of shape (N, M)
        Height data of N profiles.
    peaks, knees : arrays of shape (N,)
        X coordinates of peak point and knee point.

    Returns
    -------
    array of shape (N, 3)
        Plateau intercept, slope and boundary coordinates.

    Notes
    -----
    Plateau boundary is located by segmented regression.
    Plateau slope is forced to be nonnegative.

    See Also
    --------
    landmarks_type2 : Detects *peaks* and *knees*.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge_landmarks import landmarks_type2, plateau_type2
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     x = data.x()
    ...     Ys, Ls, _ = data[:]
    >>> lm = landmarks_type2(x, Ys, Ls, 32)
    >>> peaks, knees = lm[:, 0, 1:].T
    >>> plateau = plateau_type2(x, Ys, peaks, knees)
    >>> plateau.shape
    (22, 3)
    >>> plateau_x = np.stack([np.zeros(len(plateau)), plateau[:, 2]])
    >>> plateau_y = plateau[:, 0] + plateau_x * plateau[:, 1]
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x, Ys.T, color="gray")
    ... plt.plot(plateau_x, plateau_y)
    """
    ret = []
    for Y, peak, knee in zip(Ys, peaks, knees):
        peak, knee = np.searchsorted(x, [peak, knee])
        (b0, b1, _, psi), _ = _segreg(x[:peak], Y[:peak], x[knee])
        if b1 < 0:
            psi_idx = knee + np.argmin(np.abs(Y[knee:peak] - b0))
            b1 = 0.0
            psi = x[psi_idx]
        ret.append([b0, b1, psi])
    return np.array(ret)


def plateau_type3(x, Ys, troughs, knees):
    """Find plateau for type 3 heavy edge profiles.

    Parameters
    ----------
    x : array of shape (M,)
        X grid of profiles.
    Ys : array of shape (N, M)
        Height data of N profiles.
    troughs, knees : arrays of shape (N,)
        X coordinates of trough point and knee point.

    Returns
    -------
    array of shape (N, 3)
        Plateau intercept, slope and boundary coordinates.

    Notes
    -----
    Plateau boundary is located by segmented regression.
    Plateau slope is forced to be nonpositive.

    See Also
    --------
    landmarks_type3 : Detects *troughs* and *knees*.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge_landmarks import landmarks_type3, plateau_type3
    >>> with ProfileData(get_sample_path("Prep-Type3.h5")) as data:
    ...     x = data.x()
    ...     Ys, Ls, _ = data[:]
    >>> lm = landmarks_type3(x, Ys, Ls, 32)
    >>> troughs, knees = lm[:, 0, 2:].T
    >>> plateau = plateau_type3(x, Ys, troughs, knees)
    >>> plateau.shape
    (35, 3)
    >>> plateau_x = np.stack([np.zeros(len(plateau)), plateau[:, 2]])
    >>> plateau_y = plateau[:, 0] + plateau_x * plateau[:, 1]
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x, Ys.T, color="gray")
    ... plt.plot(plateau_x, plateau_y)
    """
    ret = []
    for Y, trough, knee in zip(Ys, troughs, knees):
        trough, knee = np.searchsorted(x, [trough, knee])
        (b0, b1, _, psi), _ = _segreg(x[:trough], Y[:trough], x[knee])
        if b1 > 0:
            psi_idx = knee + np.argmin(np.abs(Y[knee:trough] - b0))
            b1 = 0.0
            psi = x[psi_idx]
        ret.append([b0, b1, psi])
    return np.array(ret)



"""Package to locate landmarks from edge profiles."""

from .mathematical import landmarks_type2, landmarks_type3
from .plateau import plateau_type2, plateau_type3
from .preshape import dual_preshape, preshape, preshape_dual
from .pseudo import pseudo_landmarks

__all__ = [
    "pseudo_landmarks",
    "landmarks_type2",
    "landmarks_type3",
    "preshape",
    "dual_preshape",
    "preshape_dual",
    "plateau_type2",
    "plateau_type3",
]



"""Converts configuration matrices to pre-shapes.

.. note::

    Helmert sub-matrices are LRU-cached.
    The number of most recent calls can be set by the environment variable
    `HEAVYEDGE_LANDMARKS_CACHE_SIZE`, which defaults to 4.
"""

import os
import warnings
from functools import lru_cache

import numpy as np
from scipy.linalg import cho_factor, cho_solve, helmert

__all__ = [
    "preshape",
    "dual_preshape",
    "preshape_dual",
]


def preshape(Xs):
    """Convert configuration matrices to pre-shapes.

    Conversion is done using the Helmert sub-matrix.

    Parameters
    ----------
    Xs : array, shape (N, m, k)
        `N` configuration matrices of `k` landmarks in dimension `m`.

    Returns
    -------
    Zs : array, shape (N, m, k-1)
        `N` pre-shape matrices.

    See Also
    --------
    dual_preshape
        Pre-shape in configuration matrix space.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge_landmarks import pseudo_landmarks, preshape
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     x = data.x()
    ...     Ys, Ls, _ = data[:]
    >>> Zs = preshape(pseudo_landmarks(x, Ys, Ls, 10))
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(*Zs.transpose(1, 2, 0))
    """
    _, _, k = Xs.shape
    H = _helmert(k)
    HX = np.inner(Xs, H)
    scale = np.linalg.norm(HX, axis=(1, 2), keepdims=True)
    Zs = HX / scale
    return Zs


def dual_preshape(Xs):
    """Pre-shape in configuration matrix space.

    Conversion is done using the Helmert sub-matrix and its hat matrix.

    .. deprecated:: 1.2
        This function will be removed in HeavyEdge-Landmarks 2.0,
        Use :func:`preshape_dual` instead.

    Parameters
    ----------
    Xs : array, shape (N, m, k)
        `N` configuration matrices of `k` landmarks in dimension `m`.

    Returns
    -------
    Zs : array, shape (N, m, k)
        `N` pre-shape matrices.

    See Also
    --------
    preshape
        Pre-shape in its original space.

    Notes
    -----
    Because location and scale information is lost during the pre-shaping process,
    *Zs* is rank-deficient and has unit norm.
    """
    warnings.warn(
        "dual_preshape() is deprecated since HeavyEdge-Landmarks 1.2 "
        "and will be removed in 2.0. "
        "Use preshape_dual() instead.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    _, _, k = Xs.shape
    H = _helmert(k)
    HX = np.inner(Xs, H)
    scale = np.linalg.norm(HX, axis=(1, 2), keepdims=True)
    Zs = HX / scale
    hat = _helmert_hat(k)
    return np.inner(Zs, hat)


def preshape_dual(Zs):
    """Map pre-shape into configuration matrix space.

    Conversion is done using the Helmert sub-matrix and its hat matrix.

    Parameters
    ----------
    Zs : array, shape (N, m, k-1)
        `N` pre-shape matrices.

    Returns
    -------
    Zs_dual : array, shape (N, m, k)
        *Zs* in configuration matrix space.

    See Also
    --------
    preshape
        Pre-shape in its original space.

    Notes
    -----
    Because location and scale information is lost during the pre-shaping process,
    *Zs_dual* is rank-deficient and has unit norm.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge_landmarks import pseudo_landmarks, preshape, preshape_dual
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     x = data.x()
    ...     Ys, Ls, _ = data[:]
    >>> Zs = preshape(pseudo_landmarks(x, Ys, Ls, 10))
    >>> Zs_dual = preshape_dual(Zs)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(*Zs_dual.transpose(1, 2, 0))
    """
    _, _, k_minus_one = Zs.shape
    k = k_minus_one + 1
    hat = _helmert_hat(k)
    return np.inner(Zs, hat)


CACHE_SIZE = os.environ.get("HEAVYEDGE_LANDMARKS_CACHE_SIZE")
if CACHE_SIZE is not None:
    CACHE_SIZE = int(CACHE_SIZE)
else:
    CACHE_SIZE = 4


@lru_cache(maxsize=CACHE_SIZE)
def _helmert(k):
    return helmert(k)


@lru_cache(maxsize=CACHE_SIZE)
def _helmert_hat(k):
    H = helmert(k)
    # Efficient inversion (H.T @ inv(H @ H.T))
    return H.T @ cho_solve(cho_factor(H @ H.T), np.eye(H.shape[0]))



"""Detect mathematical landmarks."""

import numpy as np
from scipy.ndimage import gaussian_filter1d
from scipy.signal import find_peaks, peak_prominences

__all__ = [
    "landmarks_type2",
    "landmarks_type3",
]


def landmarks_type2(x, Ys, Ls, sigma):
    """Mathematical landmarks for type 2 heavy edge profiles.

    Type 2 heavy edge profiles have heavy edge peak.
    The following landmarks are detected:

    1. Contact point.
    2. Peak point.
    3. Knee point between plateau and peak.

    Parameters
    ----------
    x : array of shape (M,)
        X grid of profiles.
    Ys : array of shape (N, M)
        Height data of N profiles.
    Ls : array of shape (N,) and dtype=int
        Length of each profile.
    sigma : scalar
        Standard deviation of Gaussian filter for smoothing.

    Returns
    -------
    array of shape (N, 2, 3)
        X and Y coordinates of landmarks.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge_landmarks import landmarks_type2
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     x = data.x()
    ...     Ys, Ls, _ = data[:]
    >>> lm = landmarks_type2(x, Ys, Ls, 32)
    >>> lm.shape
    (22, 2, 3)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x, Ys.T, color="gray", alpha=0.5)
    ... plt.plot(*lm.transpose(1, 2, 0))
    """
    ret = []
    for Y, L in zip(Ys, Ls):
        idxs = _landmarks_type2(Y[:L], sigma)
        ret.append([x[idxs], Y[idxs]])
    return np.array(ret)


def _landmarks_type2(Y, sigma):
    cp = len(Y) - 1

    Y_smooth = gaussian_filter1d(Y, sigma)
    peaks, _ = find_peaks(Y_smooth)
    peak = peaks[-1]

    Y_ = Y_smooth[:peak]
    pts = np.column_stack([np.arange(len(Y_)), Y_])
    x, y = pts - pts[0], pts[-1] - pts[0]
    dists = x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0]
    slope = np.diff(dists)
    (extrema,) = np.nonzero(np.diff(np.sign(slope)))
    K_pos = extrema[slope[extrema] > 0]
    knee = K_pos[np.argmax(np.abs(dists[K_pos]))]

    return np.array([cp, peak, knee])


def landmarks_type3(x, Ys, Ls, sigma):
    """Mathematical landmarks for type 3 heavy edge profiles.

    Type 3 heavy edge profiles have both peak and trough.
    The following landmarks are detected:

    1. Contact point.
    2. Peak point.
    3. Trough point.
    4. Knee point between plateau and trough.

    Parameters
    ----------
    x : array of shape (M,)
        X grid of profiles.
    Ys : array of shape (N, M)
        Height data of N profiles.
    Ls : array of shape (N,) and dtype=int
        Length of each profile.
    sigma : scalar
        Standard deviation of Gaussian filter for smoothing.

    Returns
    -------
    array of shape (N, 2, 4)
        X and Y coordinates of landmarks.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge_landmarks import landmarks_type3
    >>> with ProfileData(get_sample_path("Prep-Type3.h5")) as data:
    ...     x = data.x()
    ...     Ys, Ls, _ = data[:]
    >>> lm = landmarks_type3(x, Ys, Ls, 32)
    >>> lm.shape
    (35, 2, 4)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x, Ys.T, color="gray", alpha=0.5)
    ... plt.plot(*lm.transpose(1, 2, 0))
    """
    ret = []
    for Y, L in zip(Ys, Ls):
        idxs = _landmarks_type3(Y[:L], sigma)
        ret.append([x[idxs], Y[idxs]])
    return np.array(ret)


def _landmarks_type3(Y, sigma):
    cp = len(Y) - 1

    Y_smooth = gaussian_filter1d(Y, sigma)
    peaks, _ = find_peaks(Y_smooth)
    peak = peaks[-1]

    troughs, _ = find_peaks(-Y_smooth)
    troughs = troughs[troughs < peak]

    if len(troughs) > 0:
        prominences = peak_prominences(-Y_smooth, troughs)[0]
        most_prominent_idx = np.argmax(prominences)
        trough = troughs[most_prominent_idx]

        Y_ = Y_smooth[: int(trough) + 1]
        pts = np.column_stack([np.arange(len(Y_)), Y_])
        x, y = pts - pts[0], pts[-1] - pts[0]
        dists = x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0]
        slope = np.diff(dists)
        (extrema,) = np.nonzero(np.diff(np.sign(slope)))
        K_neg = extrema[slope[extrema] < 0]
        knee = K_neg[np.argmax(np.abs(dists[K_neg]))]

    else:
        Y_ = Y_smooth[:peak]
        pts = np.column_stack([np.arange(len(Y_)), Y_])
        x, y = pts - pts[0], pts[-1] - pts[0]
        dists = x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0]
        slope = np.diff(dists)
        (extrema,) = np.nonzero(np.diff(np.sign(slope)))
        K_pos = extrema[slope[extrema] > 0]
        knee = trough = K_pos[np.argmax(np.abs(dists[K_pos]))]

    return np.array([cp, peak, trough, knee])



"""Locate pseudo-landmarks."""

import numpy as np

__all__ = [
    "pseudo_landmarks",
]


def pseudo_landmarks(x, Ys, Ls, k):
    """Sample pseudo-landmarks from edge profiles.

    Pseudo-landmarks are equidistantly sampled landmarks.

    Parameters
    ----------
    x : array of shape (M,)
        X grid of profiles.
    Ys : array of shape (N, M)
        Height data of N profiles.
    Ls : array of shape (N,) and dtype=int
        Length of each profile.
    k : int
        Number of landmarks to sample.

    Returns
    -------
    array of shape (N, 2, k)
        X and Y coordinates of landmarks.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge_landmarks import pseudo_landmarks
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     x = data.x()
    ...     Ys, Ls, _ = data[:]
    >>> lm = pseudo_landmarks(x, Ys, Ls, 10)
    >>> lm.shape
    (22, 2, 10)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x, Ys.T, color="gray", alpha=0.5)
    ... plt.plot(*lm.transpose(1, 2, 0))
    """
    ret = []
    for Y, L in zip(Ys, Ls):
        idxs = np.linspace(0, L - 1, k, dtype=int)
        ret.append([x[idxs], Y[idxs]])
    return np.array(ret)
